import 'dart:io';

import 'package:dial_videocall/src/utils/signaling.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_webrtc/flutter_webrtc.dart';

class VideoCallWidget extends StatefulWidget {
  VideoCallWidget({
    Key key, 
    @required this.server, 
    this.callerName,
    this.sessionName, 
    @required this.sessionFrom,
    this.userName, 
    this.secret, 
    this.iceServer
  }) : super(key: key);
  final String server;
  final String callerName;
  final String sessionFrom;
  final String sessionName;
  final String userName;
  final String secret;
  final String iceServer;

  @override
  _VideoCallWidgetState createState() => _VideoCallWidgetState();
}

class _VideoCallWidgetState extends State<VideoCallWidget> {

  final _localRenderer = new RTCVideoRenderer();
  final _remoteRenderer = new RTCVideoRenderer();
  bool _muted = false;
  bool _isCallIncoming = false;
  bool _isCallConnected = false;

  Signaling _signaling;
  String _callData;
  String _recordingId;

  @override
  void initState() {
    super.initState();
    initRenderers();
  }

  Future<void> initRenderers() async {
    await _localRenderer.initialize();
    await _remoteRenderer.initialize();
    _connect();
  }

  /// Hang up connected videocall.
  void _hangUp({String message}) async {
    if (_signaling != null) {
      if(_recordingId!=null) _signaling.stopRecord(_recordingId); // Record handled on web app
      if(message != null){
        await _showDialog(message);
      }
      Future.delayed(Duration.zero, () {
        Navigator.of(context).pop();
      });
    }
  }


  Future<bool> _showDialog(String message) async{
    await showDialog(
      barrierDismissible: false,
      context: context,
      builder: (context) {
        if(Platform.isAndroid){
          return AlertDialog(
            title: Text('Video Call Berakhir'),
            content: Text(message),
            actions: <Widget>[
              TextButton(
                onPressed: () {
                  Navigator.pop(context);
                },
                child: Text('Tutup')
              )
            ],
          );
        }else{
          return CupertinoAlertDialog(
            title: Text('Video Call Berakhir'),
            content: Text(message),
            actions: <Widget>[
              TextButton(
                onPressed: () {
                  Navigator.pop(context);
                },
                child: Text('Tutup')
              )
            ],
          );
        }
      }
    );
    return true;
  }

  /// Switch between front and back camera.
  void _switchCamera() {
    _signaling.switchCamera();
  }

  /// Mute the microphone.
  void _muteMic() {
    setState(() {
      _muted = !_muted;
    });
    _signaling.muteMic(_muted);
  }
  
  @override
  Widget build(BuildContext context) {
    if(_isCallConnected){
      return _videoCallUI();
    }else if (_isCallIncoming){
      return _incomingCallUI();
    }else{
      return _callConnectingUI();
    }
  }

  /// Incoming call UI.
  /// 
  /// Shown when [INCOMING_CALL] signal received.
  /// This UI consist Caller Name, Accept and Decline button.
  Widget _incomingCallUI(){
    return Scaffold(
      body: Container(
        width: MediaQuery.of(context).size.width,
        color: Colors.black,
        child: Column(
          mainAxisAlignment: MainAxisAlignment.center,
          crossAxisAlignment: CrossAxisAlignment.center,
          children: [
            Text(
              "Telepon masuk dari ${widget.callerName ?? widget.sessionFrom}",
              style: TextStyle(
                color: Colors.white,
                fontSize: 20
              ),
            ),
            SizedBox(height:50),
            Row(
              mainAxisAlignment: MainAxisAlignment.center,
              children: [
                ElevatedButton(
                  onPressed: _onAcceptCall,
                  child: Text("Accept", style: TextStyle(color: Colors.white),),
                  style: ButtonStyle(
                    backgroundColor: MaterialStateProperty.all<Color>(Colors.green),
                  ),
                ),
                SizedBox(width:30),
                ElevatedButton(
                  onPressed: _onRejectCall,
                  child: Text("Reject", style: TextStyle(color: Colors.white),),
                  style: ButtonStyle(
                    backgroundColor: MaterialStateProperty.all<Color>(Colors.red),
                  ),
                ),
              ],
            )
          ],
        ),
      ),
    );
  }

  /// Call connecting UI.
  /// 
  /// Shown when video call notification received, and web session available.
  Widget _callConnectingUI(){
    return Scaffold(
      body: Container(
        width: MediaQuery.of(context).size.width,
        color: Colors.black,
        child: Column(
          mainAxisAlignment: MainAxisAlignment.center,
          crossAxisAlignment: CrossAxisAlignment.center,
          children: [
            GestureDetector(
              child: Text(
                "Menghubungkan....",
                style: TextStyle(
                  color: Colors.white,
                  fontSize: 20
                ),
              ),
              onTap: (){
                setState(() {
                  _isCallIncoming = true;
                });
              },
            ),
            SizedBox(height:50),
            Platform.isAndroid? CircularProgressIndicator() : CupertinoActivityIndicator()
          ],
        ),
      ),
    );
  }
  /// Video Call UI.
  /// 
  /// Shown when user accepted call.
  /// Consist of remote video, local video, switch camera button, end call button and mute button.
  Widget _videoCallUI(){
    return new Scaffold(
      body: OrientationBuilder(builder: (context, orientation) {
        return new Container(
          child: new Stack(children: <Widget>[
            new Positioned(
              left: 0.0,
              right: 0.0,
              top: 0.0,
              bottom: 0.0,
              child: new Container(
                margin: new EdgeInsets.fromLTRB(0.0, 0.0, 0.0, 0.0),
                width: MediaQuery.of(context).size.width,
                height: MediaQuery.of(context).size.height,
                child: new RTCVideoView(_remoteRenderer),
                decoration: new BoxDecoration(color: Colors.black54),
              )
            ),
            new Positioned(
              left: 20.0,
              top: 40.0,
              child: new Container(
                width: orientation == Orientation.portrait ? 90.0 : 120.0,
                height: orientation == Orientation.portrait ? 120.0 : 90.0,
                child: new RTCVideoView(_localRenderer),
                decoration: new BoxDecoration(color: Colors.black54),
              ),
            ),
          ]),
        );
      }),
      floatingActionButtonLocation: FloatingActionButtonLocation.centerFloat,
      floatingActionButton: new SizedBox(
        width: 200.0,
        child: new Row(
          mainAxisAlignment: MainAxisAlignment.spaceBetween,
          children: <Widget>[
            FloatingActionButton(
              child: const Icon(Icons.switch_camera),
              onPressed: _switchCamera,
              heroTag: "btn_switchCamera",
            ),
            FloatingActionButton(
              onPressed: _hangUp,
              heroTag: "btn_hangUp",
              tooltip: 'Hangup',
              child: new Icon(Icons.call_end),
              backgroundColor: Colors.pink,
            ),
            FloatingActionButton(
              child: _muted
              ? Icon(Icons.mic)
              : Icon(Icons.mic_off),
              onPressed: _muteMic,
              heroTag: "btn_muteMic",
              backgroundColor: _muted
              ? Colors.pink
              : Colors.blue,
            )
          ]
        )
      ),
    );
  }

  @override
  void dispose() {
    super.dispose();
    _signaling?.close();
    _localRenderer?.dispose();
    _remoteRenderer?.dispose();
    print('◤◢◤◢◤◢◤◢◤◢◤ dispose() ALL ◤◢◤◢◤◢◤◢◤◢◤');
  }

  /// Accept Call
  /// 
  /// this method called when user accept the call
  void _onAcceptCall() async {
    var sessionData = widget.sessionName + "|" + _signaling.getUserID();
    _signaling.sendSignal(sessionId: sessionData, callData: _callData, type: "ACCEPTED_CALL");
    _signaling.muteMic(false);
    setState(() {
      _isCallIncoming = false;
      _isCallConnected = true;
    });
    _recordingId = await _signaling.startRecord(sessionId: widget.sessionName); // Record handled on web app
  }

  /// Reject Call
  /// 
  /// this method called when user reject the call
  void _onRejectCall(){
    _signaling.sendSignal(sessionId: widget.sessionName, callData: _callData, type: "DECLINED_CALL");
    setState(() {
      _isCallIncoming = false;
      _isCallConnected = false;
    });
    _hangUp();
  }

  /// Connect to OpenVidu
  /// 
  /// this method called after initialize local and remote renders
  /// handle any webrtc / openvidu activity from this method
  /// such as signaling, stream, etc
  Future<void> _connect() async {
    /// Check whether signaling already initialized or not
    if (_signaling == null) {

      /// Init Signaling
      _signaling = new Signaling(widget.server, widget.secret, widget.userName, widget.iceServer);

      /// Create session
      final String sessionId = await _signaling.createWebRtcSession(sessionId: widget.sessionName);
      print('◤◢◤◢◤◢◤◢◤◢◤ sessionId: $sessionId  ◤◢◤◢◤◢◤◢◤◢◤ ');

      /// Create_token
      final String token = await _signaling.createWebRtcToken(sessionId: sessionId);
      print('◤◢◤◢◤◢◤◢◤◢◤ token: $token  ◤◢◤◢◤◢◤◢◤◢◤ ');

      /// Handle signaling state change here
      _signaling.onStateChange = (SignalingState state) {
        print('_signaling.onStateChange: $state');
        switch (state) {
          case SignalingState.CallStateNew:
            break;
          case SignalingState.CallStateBye:
            break;
          case SignalingState.ConnectionError:
            _hangUp(message: "Koneksi bermasalah, video call terputus");
            break;
          case SignalingState.ConnectionClosed:
            _hangUp(message: "Koneksi video call terputus");
            break;
          default:
            break;
        }
      };

      /// Handle local stream callback here
      _signaling.onLocalStream = ((stream) {
        print('onLocalStream: ${stream.id}');
        setState(() {
          _localRenderer.srcObject = stream;
        });
      });

      /// Handle add remote stream callback here
      _signaling.onAddRemoteStream = ((stream) {
        print('onAddRemoteStream: ${stream.id}');
        setState(() {
          _remoteRenderer.srcObject = stream;
        });
      });

      /// Handle remove remote stream callback here
      _signaling.onRemoveRemoteStream = ((stream) {
        print('onRemoveRemoteStream');
        setState(() {
          _remoteRenderer.srcObject = null;
        });
      });

      /// Handle incoming signal / message callback here
      _signaling.onMessageReceive = ((stream){
        String type = stream['type'];
        String data = stream['data'];
        if(type == "CALL_REQUEST"){
          _callData = data;
          _signaling.sendSignal(sessionId: sessionId, callData: _callData, type: "CALL_REQUEST");
        }else if (type == "INCOMING_CALL"){
          setState(() {
            _isCallIncoming = true;
            _isCallConnected = false;
          });
        }else if (type == "END_CALL"){
          setState(() {
            _isCallIncoming = false;
            _isCallConnected = false;
          });
          _hangUp();
        }
      });

      /// Handle self evict callback here
      _signaling.onSelfEvict = ((userId){
        _hangUp();
      });


      /// Connect to websocket
      if(sessionId !=null && token != null){
         await _signaling.connect();
      }else{
        _hangUp(message: "Gagal menghubungkan ke server");
      }

      /// Check web session availability
      _signaling.checkSession(sessionId: widget.sessionFrom)
        .then((sessionAvailable) {
          if(sessionAvailable){
            _signaling.sendSignal(sessionId: "ready", callData: widget.sessionFrom, type: "READY_TO_CALL");
          }else{
            _hangUp(message: "Sesi video call telah berakhir");
          }
        });
    }
  }
}
